import asyncio
import logging
import random
import socket
from enum import IntEnum
from operator import attrgetter
from typing import Callable, Iterable, List, NamedTuple, Optional

import aiodns
import aiohttp

LOG = logging.getLogger("photonpump.discovery")


class NodeState(IntEnum):
    Initializing = 0
    Unknown = 1
    PreReplica = 2
    CatchingUp = 3
    Clone = 4
    Slave = 5
    PreMaster = 6
    Master = 7
    Manager = 8
    ShuttingDown = 9
    Shutdown = 10


ELIGIBLE_STATE = [NodeState.Clone, NodeState.Slave, NodeState.Master]


class NodeService(NamedTuple):
    address: str
    port: int
    secure_port: Optional[int]


class DiscoveredNode(NamedTuple):

    state: NodeState
    is_alive: bool

    internal_tcp: NodeService
    external_tcp: NodeService

    internal_http: NodeService
    external_http: NodeService


Selector = Callable[[List[DiscoveredNode]], Optional[DiscoveredNode]]


def first(elems: Iterable):
    LOG.info(elems)

    for elem in elems:
        return elem


def prefer_master(nodes: List[DiscoveredNode]) -> Optional[DiscoveredNode]:
    """
    Select the master if available, otherwise fall back to a replica.
    """
    return max(nodes, key=attrgetter("state"))


def prefer_replica(nodes: List[DiscoveredNode]) -> Optional[DiscoveredNode]:
    """
    Select a random replica if any are available or fall back to the master.
    """
    masters = [node for node in nodes if node.state == NodeState.Master]
    replicas = [node for node in nodes if node.state != NodeState.Master]

    if replicas:
        return random.choice(replicas)
    else:
        # if you have more than one master then you're on your own, bud.

        return masters[0]


def select_random(nodes: List[DiscoveredNode]) -> Optional[DiscoveredNode]:
    """
    Return a random node.
    """
    return random.choice(nodes)


def select(
    gossip: List[DiscoveredNode], selector: Optional[Selector] = None
) -> Optional[DiscoveredNode]:
    eligible_nodes = [
        node for node in gossip if node.is_alive and node.state in ELIGIBLE_STATE
    ]

    selector = selector or prefer_master
    LOG.debug(
        "Selecting node using '%s' from gossip members: %r", selector, eligible_nodes
    )

    if not eligible_nodes:
        return None

    return selector(eligible_nodes)


def read_gossip(data):
    if not data:
        LOG.debug("No gossip returned")

        return []

    LOG.debug("Received gossip for {%s} nodes", len(data["members"]))

    return [
        DiscoveredNode(
            state=NodeState[m["state"]],
            is_alive=m["isAlive"],
            internal_tcp=NodeService(m["internalTcpIp"], m["internalTcpPort"], None),
            external_tcp=NodeService(m["externalTcpIp"], m["externalTcpPort"], None),
            internal_http=NodeService(m["internalHttpIp"], m["internalHttpPort"], None),
            external_http=NodeService(m["externalHttpIp"], m["externalHttpPort"], None),
        )
        for m in data["members"]
    ]


class DiscoveryFailed(Exception):
    pass


class StaticSeedFinder:
    def __init__(self, seeds):
        self._seeds = list(seeds)
        self.candidates = []

    def reset_to_seeds(self):
        self.candidates = list(self._seeds)
        random.shuffle(self.candidates)

    def mark_failed(self, seed):
        self._seeds.remove(seed)

    async def next(self):
        if not self.candidates:
            self.reset_to_seeds()

        if not self.candidates:
            return None

        return self.candidates.pop()

    def add_node(self, node):
        self.candidates.append(node)


class DnsSeedFinder:
    def __init__(self, name, resolver, port=2113):
        self.name = name
        self.resolver = resolver
        self.port = port
        self.seeds = []
        self.failed_node = NodeService("_", 0, None)

    async def reset_to_dns(self):
        max_attempt = 100
        current_attempt = 0

        while current_attempt < max_attempt:
            LOG.info(
                "Attempting to discover gossip nodes from DNS name %s; "
                "attempt %d of %d",
                self.name,
                current_attempt,
                max_attempt,
            )
            try:
                result = await self.resolver.query(self.name, "A")
                random.shuffle(result)

                if result:
                    LOG.debug("Found %s hosts for name %s", len(result), self.name)
                    current_attempt = 0
                    self.seeds = [
                        NodeService(address=node.host, port=self.port, secure_port=None)
                        for node in result
                        if node.host != self.failed_node.address
                    ]

                    break
            except aiodns.error.DNSError:
                LOG.warning(
                    "Failed to fetch gossip seeds for dns name %s",
                    self.name,
                    exc_info=True,
                )
            current_attempt += 1
            await asyncio.sleep(1)

    def mark_failed(self, seed):
        if seed in self.seeds:
            self.seeds.remove(seed)
        self.failed_node = seed

    async def next(self):
        if not self.seeds:
            await self.reset_to_dns()

        if not self.seeds:
            return None

        return self.seeds[0]

    def add_node(self, node):
        self.seeds.append(node)


async def fetch_new_gossip(session, seed):
    if not seed:
        return []

    LOG.debug("Fetching gossip from http://%s:%s/gossip", seed.address, seed.port)
    try:
        resp = await session.get(f"http://{seed.address}:{seed.port}/gossip")
        data = await resp.json()

        return read_gossip(data)
    except aiohttp.ClientError:
        LOG.exception(
            "Failed loading gossip from http://%s:%s/gossip", seed.address, seed.port
        )

        return None


class SingleNodeDiscovery:
    def __init__(self, node):
        self.node = node
        self.failed = False

    def mark_failed(self, node):
        if node == self.node:
            self.failed = True

    async def discover(self):
        if self.failed:
            raise DiscoveryFailed()
        LOG.debug("SingleNodeDiscovery returning node %s", self.node)

        return self.node


class DiscoveryStats(NamedTuple):

    node: NodeService
    attempts: int
    successes: int
    failures: int
    consecutive_failures: int


class Stats(dict):
    def __missing__(self, key):
        value = self[key] = DiscoveryStats(key, 0, 0, 0, 0)

        return value

    def record_success(self, node):
        val = self[node]
        self[node] = val._replace(
            attempts=(val.attempts + 1),
            successes=(val.successes + 1),
            consecutive_failures=0,
        )

    def record_failure(self, node):
        val = self[node]
        self[node] = val._replace(
            attempts=(val.attempts + 1),
            failures=(val.failures + 1),
            consecutive_failures=(val.consecutive_failures + 1),
        )


class ClusterDiscovery:
    def __init__(self, seed_finder, http_session, retry_policy, selector):
        self.session = http_session
        self.seeds = seed_finder
        self.last_gossip = []
        self.best_node = None
        self.retry_policy = retry_policy
        self.selector = selector

    def close(self):
        self.session.close()

    def mark_failed(self, node):
        self.seeds.mark_failed(node)

    def record_gossip(self, node, gossip):
        self.last_gossip = gossip

        for member in gossip:
            self.seeds.add_node(member.external_http)
        self.best_node = select(gossip, self.selector)
        self.retry_policy.record_success(node)

    async def get_gossip(self):
        while True:

            seed = await self.seeds.next()
            LOG.info(f"Found gossip seed {seed}")

            if not seed:
                raise DiscoveryFailed()

            await self.retry_policy.wait(seed)
            gossip = await fetch_new_gossip(self.session, seed)

            if gossip:
                self.record_gossip(seed, gossip)

                return gossip
            else:
                self.retry_policy.record_failure(seed)

                if not self.retry_policy.should_retry(seed):
                    self.seeds.mark_failed(seed)

    async def discover(self):
        gossip = await self.get_gossip()

        if gossip:
            if self.best_node:
                return self.best_node.external_tcp
        raise DiscoveryFailed()


class DiscoveryRetryPolicy:
    def __init__(
        self,
        retries_per_node=3,
        retry_interval=0.5,
        jitter=0.5,
        multiplier=1.5,
        max_interval=60,
    ):
        self.stats = Stats()
        self.retries_per_node = retries_per_node
        self.retry_interval = retry_interval
        self.jitter = jitter
        self.multiplier = multiplier
        self.max_interval = max_interval
        self.next_interval = self.retry_interval

    def should_retry(self, node):
        stats = self.stats[node]
        return (
            self.retries_per_node == 0
            or stats.consecutive_failures < self.retries_per_node
        )

    async def wait(self, seed):
        stats = self.stats[seed]

        if stats.consecutive_failures == 0:
            return

        next_interval = (
            self.retry_interval * self.multiplier * stats.consecutive_failures
        )
        maxinterval = next_interval + self.jitter
        mininterval = next_interval - self.jitter
        interval = random.uniform(mininterval, maxinterval)

        LOG.debug(f"Discovery retry policy sleeping for {interval} secs")
        await asyncio.sleep(interval)

    def record_success(self, node):
        self.stats.record_success(node)
        self.next_interval = self.retry_interval

    def record_failure(self, node):
        self.stats.record_failure(node)


def get_discoverer(
    host, port, discovery_host, discovery_port, selector: Optional[Selector] = None
):
    if discovery_host is None:
        LOG.info("Using single-node discoverer")

        return SingleNodeDiscovery(NodeService(host or "localhost", port, None))

    session = aiohttp.ClientSession()
    try:
        socket.inet_aton(discovery_host)
        LOG.info("Using cluster node discovery with a static seed")

        return ClusterDiscovery(
            StaticSeedFinder([NodeService(discovery_host, discovery_port, None)]),
            session,
            DiscoveryRetryPolicy(),
            selector,
        )
    except socket.error:
        LOG.info("Using cluster node discovery with DNS")
        resolver = aiodns.DNSResolver()

        return ClusterDiscovery(
            DnsSeedFinder(discovery_host, resolver, discovery_port),
            session,
            DiscoveryRetryPolicy(),
            selector,
        )



import datetime
import json
import math
import struct
from collections import namedtuple
from enum import IntEnum
from typing import Any, Dict, Sequence, NamedTuple, Optional
from uuid import UUID, uuid4

from . import messages_pb2

HEADER_LENGTH = 1 + 1 + 16
SIZE_UINT_32 = 4
_LENGTH = struct.Struct("<I")
_HEAD = struct.Struct("<BBIIII")

ROUND_ROBIN = "RoundRobin"
DISPATCH_TO_SINGLE = "DisptchToSingle"
PINNED = "Pinned"


def make_enum(descriptor):
    vals = [(x.name, x.number) for x in descriptor.values]

    return IntEnum(descriptor.name, vals)


class TcpCommand(IntEnum):

    HeartbeatRequest = 0x01
    HeartbeatResponse = 0x02

    Ping = 0x03
    Pong = 0x04

    WriteEvents = 0x82
    WriteEventsCompleted = 0x83

    Read = 0xB0
    ReadEventCompleted = 0xB1
    ReadStreamEventsForward = 0xB2
    ReadStreamEventsForwardCompleted = 0xB3
    ReadStreamEventsBackward = 0xB4
    ReadStreamEventsBackwardCompleted = 0xB5
    ReadAllEventsForward = 0xB6
    ReadAllEventsForwardCompleted = 0xB7
    ReadAllEventsBackward = 0xB8
    ReadAllEventsBackwardCompleted = 0xB9

    SubscribeToStream = 0xC0
    SubscriptionConfirmation = 0xC1
    StreamEventAppeared = 0xC2
    UnsubscribeFromStream = 0xC3
    SubscriptionDropped = 0xC4

    ConnectToPersistentSubscription = 0xC5
    PersistentSubscriptionConfirmation = 0xC6
    PersistentSubscriptionStreamEventAppeared = 0xC7
    CreatePersistentSubscription = 0xC8
    CreatePersistentSubscriptionCompleted = 0xC9
    DeletePersistentSubscription = 0xCA
    DeletePersistentSubscriptionCompleted = 0xCB
    PersistentSubscriptionAckEvents = 0xCC
    PersistentSubscriptionNakEvents = 0xCD
    UpdatePersistentSubscription = 0xCE
    UpdatePersistentSubscriptionCompleted = 0xCF

    BadRequest = 0xF0
    NotHandled = 0xF1
    Authenticate = 0xF2
    Authenticated = 0xF3
    NotAuthenticated = 0xF4
    IdentifyClient = 0xF5
    ClientIdentified = 0xF6


class StreamDirection(IntEnum):
    Forward = 0
    Backward = 1


class ContentType(IntEnum):
    Json = 0x01
    Binary = 0x00


class OperationFlags(IntEnum):
    Empty = 0x00
    Authenticated = 0x01


class Position(NamedTuple):

    commit: int
    prepare: int

    @classmethod
    def for_direction(cls, direction, pos):
        if pos is None:
            return (
                Position(0, 0)
                if direction == StreamDirection.Forward
                else Position(-1, -1)
            )
        elif pos is Beginning:
            return Position(0, 0)
        elif pos is End:
            return Position(-1, -1)

        return pos


Position.min = Position(-1, -1)


class _PositionSentinel:
    """
    Provides magic values for beginning/end
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Position: " + self.name


Beginning = _PositionSentinel("Beginning")
End = _PositionSentinel("End")

OperationResult = make_enum(messages_pb2._OPERATIONRESULT)
NotHandledReason = make_enum(messages_pb2._NOTHANDLED_NOTHANDLEDREASON)
SubscriptionDropReason = make_enum(
    messages_pb2._SUBSCRIPTIONDROPPED_SUBSCRIPTIONDROPREASON
)


class Credential:
    def __init__(self, username, password):
        self.username = username
        self.password = password

        username_bytes = username.encode("UTF-8")
        password_bytes = password.encode("UTF-8")

        self.length = 2 + len(username_bytes) + len(password_bytes)
        self.bytes = bytearray()

        self.bytes.extend(len(username_bytes).to_bytes(1, byteorder="big"))
        self.bytes.extend(username_bytes)
        self.bytes.extend(len(password_bytes).to_bytes(1, byteorder="big"))
        self.bytes.extend(password_bytes)

    @classmethod
    def from_bytes(cls, data):
        """
        I am so sorry.
        """
        len_username = int.from_bytes(data[0:2], byteorder="big")
        offset_username = 2 + len_username
        username = data[2:offset_username].decode("UTF-8")
        offset_password = 2 + offset_username
        len_password = int.from_bytes(
            data[offset_username:offset_password], byteorder="big"
        )
        pass_begin = offset_password
        pass_end = offset_password + len_password
        password = data[pass_begin:pass_end].decode("UTF-8")

        return cls(username, password)


class InboundMessage:
    def __init__(
        self, conversation_id: UUID, command: TcpCommand, payload: bytes = None
    ) -> None:
        self.conversation_id = conversation_id
        self.command = command
        self.payload = payload or b""
        self.data_length = len(payload)
        self.length = HEADER_LENGTH + self.data_length

    @property
    def header_bytes(self):
        data = bytearray(SIZE_UINT_32 + 2)
        struct.pack_into("<IBB", data, 0, self.length, self.command, 0)
        data.extend(self.conversation_id.bytes_le)

        return data

    def __repr__(self):
        return self.__str__() + "\n" + dump(self.header_bytes, self.payload)

    def __str__(self):
        return "%s (%s) of %s flags=%d" % (
            TcpCommand(self.command).name,
            self.conversation_id,
            sizeof_fmt(self.length),
            0,
        )


class OutboundMessage:
    def __init__(
        self,
        conversation_id: UUID,
        command: TcpCommand,
        payload: Any,
        creds: Credential = None,
        one_way: bool = False,
        require_master=False,
    ) -> None:
        self.conversation_id = conversation_id
        self.command = command
        self.payload = payload
        self.creds = creds
        self.require_master = require_master

        self.data_length = len(payload)
        self.length = HEADER_LENGTH + self.data_length

        if self.creds:
            self.length += creds.length
        self.one_way = one_way

    @property
    def header_bytes(self):
        data = bytearray(SIZE_UINT_32 + 2)
        struct.pack_into(
            "<IBB",
            data,
            1 if self.require_master else 0,
            self.length,
            self.command.value,
            1 if self.creds else 0,
        )
        data.extend(self.conversation_id.bytes_le)

        if self.creds:
            data.extend(self.creds.bytes)

        return data

    def __repr__(self):
        return dump(self.header_bytes, self.payload)

    def __str__(self):
        return "%s (%s) of %s flags=%d" % (
            TcpCommand(self.command).name,
            self.conversation_id,
            sizeof_fmt(self.length),
            0,
        )

    def __eq__(self, other):
        return isinstance(other, OutboundMessage) and repr(self) == repr(other)


class ExpectedVersion(IntEnum):
    """Static values for concurrency control

    Attributes:
        Any: No concurrency control.
        StreamMustNotExist: The request should fail if the stream
          already exists.
        StreamMustBeEmpty: The request should fail if the stream
          does not exist, or if the stream already contains events.
        StreamMustExist: The request should fail if the stream
          does not exist.
    """

    Any = -2
    StreamMustNotExist = -1
    StreamMustBeEmpty = 0
    StreamMustExist = -4


JsonDict = Dict[str, Any]
Header = namedtuple(
    "photonpump_result_header",
    ["size", "cmd", "flags", "correlation_id", "username", "password"],
)


def parse_header(length: bytearray, data: bytearray) -> Header:
    (size,) = _LENGTH.unpack(length)

    return Header(size, data[0], data[1], UUID(bytes_le=data[2:18]), None, None)


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0

    return "%.1f%s%s" % (num, "Yi", suffix)


def print_header(header):
    return "%s (%s) of %s flags=%d" % (
        TcpCommand(header.cmd).name,
        header.correlation_id,
        sizeof_fmt(header.size),
        header.flags,
    )


Header.__repr__ = print_header
Header.__new__.__defaults__ = (None, None)

NewEventData = namedtuple("photonpump_event", ["id", "type", "data", "metadata"])

EventRecord = namedtuple(
    "photonpump_eventrecord",
    ["stream", "id", "event_number", "type", "data", "metadata", "created"],
)


def _json(self):
    return json.loads(self.data.decode("UTF-8"))


EventRecord.json = _json


class Event:
    def __init__(
        self, event: EventRecord, link: EventRecord, position: Optional[Position] = None
    ) -> None:
        self.event = event
        self.link = link
        self.stream = event.stream
        self.id = event.id
        self.event_number = self.received_event.event_number
        self.type = event.type
        self.data = event.data
        self.metadata = event.metadata
        self.created = event.created
        self.position = position

    @property
    def received_event(self) -> EventRecord:
        return self.link or self.event

    def json(self):
        return json.loads(self.data.decode("UTF-8"))

    def __repr__(self):
        return f"<Event {self.created}: {self.event_number}@{self.stream}:{self.type}>"


class StreamSlice(list):
    def __init__(
        self,
        events: Sequence[Event],
        next_event_number: int,
        last_event_number: int,
        prepare_position: int = None,
        commit_position: int = None,
        is_end_of_stream: bool = False,
    ) -> None:
        super().__init__(events)
        self.next_event_number = next_event_number
        self.last_event_number = last_event_number
        self.prepare_position = prepare_position
        self.commit_position = commit_position
        self.is_end_of_stream = is_end_of_stream
        self.events = events


class AllStreamSlice(list):
    def __init__(
        self, events: Sequence[Event], position: Position, next_position: Position
    ) -> None:
        super().__init__(events)
        self.next_position = next_position
        self.start_position = position
        self.events = events


def dump(*chunks: bytearray):
    data = bytearray()

    for chunk in chunks:
        data.extend(chunk)

    length = len(data)
    rows = length / 16

    dump = []

    for i in range(0, math.ceil(rows)):
        offset = i * 16
        row = data[offset : offset + 16]
        hex = "{0: <47}".format(" ".join("{:02x}".format(x) for x in row))
        dump.append("%06d: %s | %a" % (offset, hex, bytes(row)))

    return "\n" + "\n".join(dump)


def _make_event(record: messages_pb2.ResolvedEvent):

    if type(record) == messages_pb2.ResolvedEvent:
        position = Position(record.commit_position, record.prepare_position)
    else:
        position = None

    link = (
        EventRecord(
            record.link.event_stream_id,
            UUID(bytes_le=bytes(record.link.event_id)),
            record.link.event_number,
            record.link.event_type,
            record.link.data,
            record.link.metadata,
            datetime.datetime.fromtimestamp(record.link.created_epoch / 1e3),
        )
        if record.HasField("link")
        else None
    )

    if not record.HasField("event"):
        return Event(link, None)

    event = EventRecord(
        record.event.event_stream_id,
        UUID(bytes_le=bytes(record.event.event_id)),
        record.event.event_number,
        record.event.event_type,
        record.event.data,
        record.event.metadata,
        datetime.datetime.fromtimestamp(record.event.created_epoch / 1e3),
    )

    return Event(event, link, position)


def NewEvent(
    type: str, id: UUID = None, data: JsonDict = None, metadata: JsonDict = None
) -> NewEventData:
    """Build the data structure for a new event.

    Args:
        type: An event type.
        id: The uuid identifier for the event.
        data: A dict containing data for the event. These data
            must be json serializable.
        metadata: A dict containing metadata about the event.
            These must be json serializable.
    """

    return NewEventData(id or uuid4(), type, data, metadata)


ReadEventResult = make_enum(messages_pb2._READEVENTCOMPLETED_READEVENTRESULT)

ReadStreamResult = make_enum(messages_pb2._READSTREAMEVENTSCOMPLETED_READSTREAMRESULT)

ReadAllResult = make_enum(messages_pb2._READALLEVENTSCOMPLETED_READALLRESULT)

SubscriptionResult = make_enum(
    messages_pb2._CREATEPERSISTENTSUBSCRIPTIONCOMPLETED_CREATEPERSISTENTSUBSCRIPTIONRESULT
)


class SubscriptionCreatedResponse:
    def __init__(self, result: SubscriptionResult, reason: str) -> None:
        self.reason = reason
        self.result = result
